@@ -115,6 +115,12 @@ typedef union
 
 extern DItype __fixunssfdi (SFtype a);
 extern DItype __fixunsdfdi (DFtype a);
+#if LONG_DOUBLE_TYPE_SIZE == 96
+extern DItype __fixunsxfdi (XFtype a);
+#endif
+#if LONG_DOUBLE_TYPE_SIZE == 128
+extern DItype __fixunstfdi (TFtype a);
+#endif
 
 #if defined (L_negdi2) || defined (L_divdi3) || defined (L_moddi3)
 #if defined (L_divdi3) || defined (L_moddi3)
@@ -266,28 +272,6 @@ __ashrdi3 (u, b)
 #endif
 
 #ifdef L_ffsdi2
-
-#ifdef foo__i386__
-
-#undef	ffs
-
-__inline__ static word_type
-ffs (word_type x)
-{
-  word_type cnt;
-  word_type tmp;
-
-  asm ("xorl %0,%0\n"		/* Set CNT to zero.  */
-       "bsfl %2,%1\n"		/* Count low bits in X and store in %1.  */
-       "jz 1f\n"		/* Jump if OK, i.e. X was non-zero.  */
-       "leal 1(%3),%0\n"	/* Return bsfl-result plus one on %0.  */
-       "1:" : "=a" (cnt), "=r" (tmp) : "rm" (x), "1" (tmp));
-
-  return cnt;
-}
-
-#endif
-
 DItype
 __ffsdi2 (u)
      DItype u;
@@ -662,6 +646,7 @@ __udivmoddi4 (n, d, rp)
 
 #ifdef L_divdi3
 UDItype __udivmoddi4 ();
+
 DItype
 __divdi3 (u, v)
      DItype u, v;
@@ -1204,6 +1189,73 @@ asm ("___builtin_saveregs:");
 	asm ("	st.l	%r28,12(%r16)"); /* pointer to overflow args */
 
 #else /* not __svr4__ */
+#if defined(__PARAGON__)
+	/*
+	 *	we'll use SVR4-ish varargs but need SVR3.2 assembler syntax,
+	 *	and we stand a better chance of hooking into libraries
+	 *	compiled by PGI.  [andyp@ssd.intel.com]
+	 */
+	asm ("	.text");
+	asm ("	.align	4");
+	asm (".globl	__builtin_saveregs");
+asm ("__builtin_saveregs:");
+	asm (".globl	___builtin_saveregs");
+asm ("___builtin_saveregs:");
+
+        asm ("	andnot	0x0f,sp,sp");	/* round down to 16-byte boundary */
+	asm ("	adds	-96,sp,sp");	/* allocate stack space for reg save
+					   area and also for a new va_list
+					   structure */
+	/* Save all argument registers in the arg reg save area.  The
+	   arg reg save area must have the following layout (according
+	   to the svr4 ABI):
+
+		struct {
+		  union  {
+		    float freg[8];
+		    double dreg[4];
+		  } float_regs;
+		  long	ireg[12];
+		};
+	*/
+
+	asm ("	fst.q	f8,  0(sp)");
+	asm ("	fst.q	f12,16(sp)"); 
+	asm ("	st.l	r16,32(sp)");
+	asm ("	st.l	r17,36(sp)"); 
+	asm ("	st.l	r18,40(sp)");
+	asm ("	st.l	r19,44(sp)");
+	asm ("	st.l	r20,48(sp)");
+	asm ("	st.l	r21,52(sp)");
+	asm ("	st.l	r22,56(sp)");
+	asm ("	st.l	r23,60(sp)");
+	asm ("	st.l	r24,64(sp)");
+	asm ("	st.l	r25,68(sp)");
+	asm ("	st.l	r26,72(sp)");
+	asm ("	st.l	r27,76(sp)");
+
+	asm ("	adds	80,sp,r16");  /* compute the address of the new
+					   va_list structure.  Put in into
+					   r16 so that it will be returned
+					   to the caller.  */
+
+	/* Initialize all fields of the new va_list structure.  This
+	   structure looks like:
+
+		typedef struct {
+		    unsigned long	ireg_used;
+		    unsigned long	freg_used;
+		    long		*reg_base;
+		    long		*mem_ptr;
+		} va_list;
+	*/
+
+	asm ("	st.l	r0, 0(r16)"); /* nfixed */
+	asm ("	st.l	r0, 4(r16)"); /* nfloating */
+	asm ("  st.l    sp, 8(r16)"); /* __va_ctl points to __va_struct.  */
+	asm ("	bri	r1");		/* delayed return */
+	asm ("	 st.l	r28,12(r16)"); /* pointer to overflow args */
+#else /* not __PARAGON__ */
 	asm ("	.text");
 	asm ("	.align	4");
 
@@ -1241,6 +1293,7 @@ asm ("___builtin_saveregs:");
 	asm ("	mov	r30,sp");
 				/* recover stack and pass address to start 
 				   of data.  */
+#endif /* not __PARAGON__ */
 #endif /* not __svr4__ */
 #else /* not __i860__ */
 #ifdef __sparc__
@@ -1307,35 +1360,202 @@ __eprintf (string, expression, line, filename)
 #endif
 
 #ifdef L_bb
-/* Avoid warning from ranlib about empty object file.  */
-void
-__bb_avoid_warning ()
-{}
 
-#if defined (__sun__) && defined (__mc68000__)
+/* Structure emitted by -a  */
 struct bb
 {
-  int initialized;
-  char *filename;
-  int *counts;
-  int ncounts;
-  int zero_word;
-  int *addresses;
+  long zero_word;
+  const char *filename;
+  long *counts;
+  long ncounts;
+  struct bb *next;
+  const unsigned long *addresses;
+
+  /* Older GCC's did not emit these fields.  */
+  long nwords;
+  const char **functions;
+  const long *line_nums;
+  const char **filenames;
 };
 
-extern int ___tcov_init;
+#ifdef BLOCK_PROFILER_CODE
+BLOCK_PROFILER_CODE
+#else
+#ifndef inhibit_libc
 
-__bb_init_func (blocks)
-	struct bb *blocks;
+/* Simple minded basic block profiling output dumper for
+   systems that don't provde tcov support.  At present,
+   it requires atexit and stdio.  */
+
+#undef NULL /* Avoid errors if stdio.h and our stddef.h mismatch.  */
+#include <stdio.h>
+
+#ifdef HAVE_ATEXIT
+extern void atexit (void (*) (void));
+#define ON_EXIT(FUNC,ARG) atexit ((FUNC))
+#else
+#ifdef sun
+extern void on_exit (void*, void*);
+#define ON_EXIT(FUNC,ARG) on_exit ((FUNC), (ARG))
+#endif
+#endif
+
+static struct bb *bb_head = (struct bb *)0;
+
+/* Return the number of digits needed to print a value */
+/* __inline__ */ static int num_digits (long value, int base)
 {
-  if (! ___tcov_init)
-    ___tcov_init_func ();
+  int minus = (value < 0 && base != 16);
+  unsigned long v = (minus) ? -value : value;
+  int ret = minus;
 
-  ___bb_link (blocks->filename, blocks->counts, blocks->ncounts);
+  do
+    {
+      v /= base;
+      ret++;
+    }
+  while (v);
+
+  return ret;
 }
 
+void
+__bb_exit_func (void)
+{
+  FILE *file = fopen ("bb.out", "a");
+  long time_value;
+
+  if (!file)
+    perror ("bb.out");
+
+  else
+    {
+      struct bb *ptr;
+
+      /* This is somewhat type incorrect, but it avoids worrying about
+	 exactly where time.h is included from.  It should be ok unless
+	 a void * differs from other pointer formats, or if sizeof(long)
+	 is < sizeof (time_t).  It would be nice if we could assume the
+	 use of rationale standards here.  */
+
+      time((void *) &time_value);
+      fprintf (file, "Basic block profiling finished on %s\n", ctime ((void *) &time_value));
+
+      /* We check the length field explicitly in order to allow compatibility
+	 with older GCC's which did not provide it.  */
+
+      for (ptr = bb_head; ptr != (struct bb *)0; ptr = ptr->next)
+	{
+	  int i;
+	  int func_p	= (ptr->nwords >= sizeof (struct bb) && ptr->nwords <= 1000);
+	  int line_p	= (func_p && ptr->line_nums);
+	  int file_p	= (func_p && ptr->filenames);
+	  long ncounts	= ptr->ncounts;
+	  long cnt_max  = 0;
+	  long line_max = 0;
+	  long addr_max = 0;
+	  int file_len	= 0;
+	  int func_len	= 0;
+	  int blk_len	= num_digits (ncounts, 10);
+	  int cnt_len;
+	  int line_len;
+	  int addr_len;
+
+	  fprintf (file, "File %s, %ld basic blocks \n\n",
+		   ptr->filename, ncounts);
+
+	  /* Get max values for each field.  */
+	  for (i = 0; i < ncounts; i++)
+	    {
+	      const char *p;
+	      int len;
+
+	      if (cnt_max < ptr->counts[i])
+		cnt_max = ptr->counts[i];
+
+	      if (addr_max < ptr->addresses[i])
+		addr_max = ptr->addresses[i];
+
+	      if (line_p && line_max < ptr->line_nums[i])
+		line_max = ptr->line_nums[i];
+
+	      if (func_p)
+		{
+		  p = (ptr->functions[i]) ? (ptr->functions[i]) : "<none>";
+		  len = strlen (p);
+		  if (func_len < len)
+		    func_len = len;
+		}
+
+	      if (file_p)
+		{
+		  p = (ptr->filenames[i]) ? (ptr->filenames[i]) : "<none>";
+		  len = strlen (p);
+		  if (file_len < len)
+		    file_len = len;
+		}
+	    }
+
+	  addr_len = num_digits (addr_max, 16);
+	  cnt_len  = num_digits (cnt_max, 10);
+	  line_len = num_digits (line_max, 10);
+
+	  /* Now print out the basic block information.  */
+	  for (i = 0; i < ncounts; i++)
+	    {
+	      fprintf (file,
+		       "    Block #%*d: executed %*ld time(s) address= 0x%.*lx",
+		       blk_len, i+1,
+		       cnt_len, ptr->counts[i],
+		       addr_len, ptr->addresses[i]);
+
+	      if (func_p)
+		fprintf (file, " function= %-*s", func_len,
+			 (ptr->functions[i]) ? ptr->functions[i] : "<none>");
+
+	      if (line_p)
+		fprintf (file, " line= %*d", line_len, ptr->line_nums[i]);
+
+	      if (file_p)
+		fprintf (file, " file= %s",
+			 (ptr->filenames[i]) ? ptr->filenames[i] : "<none>");
+
+	      fprintf (file, "\n");
+	    }
+
+	  fprintf (file, "\n");
+	  fflush (file);
+	}
+
+      fprintf (file, "\n\n");
+      fclose (file);
+    }
+}
+
+void
+__bb_init_func (struct bb *blocks)
+{
+  /* User is supposed to check whether the first word is non-0,
+     but just in case.... */
+
+  if (blocks->zero_word)
+    return;
+
+#ifdef ON_EXIT
+  /* Initialize destructor.  */
+  if (!bb_head)
+    ON_EXIT (__bb_exit_func, 0);
 #endif
-#endif
+
+  /* Set up linked list.  */
+  blocks->zero_word = 1;
+  blocks->next = bb_head;
+  bb_head = blocks;
+}
+
+#endif /* not inhibit_libc */
+#endif /* not BLOCK_PROFILER_CODE */
+#endif /* L_bb */
 
 /* frills for C++ */
 
@@ -1447,6 +1667,9 @@ void
 __clear_cache (beg, end)
      char *beg, *end;
 {
+#ifdef CLEAR_INSN_CACHE 
+  CLEAR_INSN_CACHE (beg, end);
+#else
 #ifdef INSN_CACHE_SIZE
   static char array[INSN_CACHE_SIZE + INSN_CACHE_PLANE_SIZE + INSN_CACHE_LINE_WIDTH];
   static int initialized = 0;
@@ -1545,6 +1768,7 @@ __clear_cache (beg, end)
 #endif /* just one plane */
 #endif /* Cache is large */
 #endif /* Cache exists */
+#endif /* CLEAR_INSN_CACHE */
 }
 
 #endif /* L_clear_cache */
@@ -1557,6 +1781,40 @@ __clear_cache (beg, end)
 TRANSFER_FROM_TRAMPOLINE 
 #endif
 
+#if defined (NeXT) && defined (__MACH__)
+
+/* Make stack executable so we can call trampolines on stack.
+   This is called from INITIALIZE_TRAMPOLINE in next.h.  */
+
+#include <mach/mach.h>
+
+void
+__enable_execute_stack (addr)
+     char *addr;
+{
+  kern_return_t r;
+  char *eaddr = addr + TRAMPOLINE_SIZE;
+  vm_address_t a = (vm_address_t) addr;
+
+  /* turn on execute access on stack */
+  r = vm_protect (task_self (), a, TRAMPOLINE_SIZE, FALSE, VM_PROT_ALL);
+  if (r != KERN_SUCCESS)
+    {
+      mach_error("vm_protect VM_PROT_ALL", r);
+      exit(1);
+    }
+
+  /* We inline the i-cache invalidation for speed */
+
+#ifdef CLEAR_INSN_CACHE
+  CLEAR_INSN_CACHE (addr, eaddr);
+#else
+  __clear_cache ((int) addr, (int) eaddr);
+#endif
+} 
+
+#endif /* defined (NeXT) && defined (__MACH__) */
+
 #ifdef __convex__
 
 /* Make stack executable so we can call trampolines on stack.
@@ -1615,6 +1873,14 @@ __enable_execute_stack ()
 #ifdef L__main
 
 #include "gbl-ctors.h"
+/* Some systems use __main in a way incompatible with its use in gcc, in these
+   cases use the macros NAME__MAIN to give a quoted symbol and SYMBOL__MAIN to
+   give the same symbol without quotes for an alternative entry point.  You
+   must define both, or niether. */
+#ifndef NAME__MAIN
+#define NAME__MAIN "__main"
+#define SYMBOL__MAIN __main
+#endif
 
 /* Run all the global destructors on exit from the program.  */
 
@@ -1674,7 +1940,7 @@ __do_global_ctors ()
    to run __do_global_ctors, so we need not do anything here.  */
 
 void
-__main ()
+SYMBOL__MAIN ()
 {
   /* Support recursive calls to `main': run initializers just once.  */
   static int initialized = 0;
@@ -1723,7 +1989,7 @@ func_ptr __DTOR_LIST__[2];
 
 extern void __do_global_dtors ();
 extern void _cleanup ();
-extern volatile void _exit ();
+extern void _exit () __attribute__ ((noreturn));
 
 void 
 exit (status)
