@@ -28,6 +28,16 @@ static char sccsid[] = "@(#)gethostnamadr.c	6.39 (Berkeley) 1/4/90";
 
 #define _PATH_HOSTCONF	__PATH_ETC_INET"/host.conf"
 
+/* Strict but flexible checking in _PATH_HOSTCONF file.
+
+ * (+) Keywords need not start at the begining of the line.
+ * (+) Invalid keywords are warned.
+ * (+) Invalid booleans (other than "on" or "off") are warned.
+
+ * Mitch DSouza <m.dsouza@mrc-apu.cam.ac.uk> - 30th October 1993
+ */
+#define STRICT
+
 #define SERVICE_NONE	0
 #define SERVICE_BIND	1
 #define SERVICE_HOSTS	2
@@ -107,14 +117,18 @@ extern int h_errno;
 #endif
 extern int errno;
 
+#if NLS
+#include "../nls/nl_types.h"
+#endif
+
 static void
 dotrimdomain(c)
 char *c;
 {
-	/* assume c points to the start of a host name; trim off any 
+	/* assume c points to the start of a host name; trim off any
 	   domain name matching any of the trimdomains */
 	int d,l1,l2;
-	
+
 	for(d=0;d<numtrimdomains;d++){
 		l1=strlen(trimdomain[d]);
 		l2=strlen(c);
@@ -136,8 +150,17 @@ struct hostent *h;
 	return(h);
 }
 
+
+#if NLS
+#define bad_config_format(cmd) \
+    fprintf(stderr, catgets(_libc_cat, NetMiscSet, \
+		NetMiscResolvIncorrectFormat, \
+		"resolv+: %s: \"%s\" command incorrectly formatted.\n"), \
+		hostconf, cmd);
+#else
 #define bad_config_format(cmd) \
     fprintf(stderr, "resolv+: %s: \"%s\" command incorrectly formatted.\n", hostconf, cmd);
+#endif
 
 /* static void debug_breakout(x)
 long x;
@@ -196,7 +219,7 @@ struct hostent *h;
 			numitab is 0  : should never happen.
 			numitab is 1,2,3,... :  number of valid entries in the table.
 	***/
-	if (!numitab) return;	/* no entries in table */
+	if (!numitab || h==NULL) return;	/* no entries in table */
 	if (numitab==-1) { /* build the table */
 		int fd, err;
 		struct ifconf ifs;
@@ -251,7 +274,7 @@ struct hostent *h;
 	}
 
 	/**** loop through table for each (address,interface) combo ****/
-	for ( r = (struct in_addr **) (h->h_addr_list); *r; r++) {	/* loop through the addresses */
+	for ( r = (struct in_addr **) (h->h_addr_list); r && *r; r++) {	/* loop through the addresses */
 		for ( itp = itab, cnt=numitab; cnt; itp++,cnt--) {	/* loop though the interfaces */
 			if (( (*r)->s_addr & itp->netmask) == itp->address) {	/* compare */
 				/* We found a match.  Swap it into [0] */
@@ -274,6 +297,9 @@ init_services()
 	char *tdp = trimdomainbuf;
 	char *hostconf;
 
+#if NLS
+	libc_nls_init();
+#endif
 	if(NULL==(hostconf=getenv(ENV_HOSTCONF))){
 		hostconf=_PATH_HOSTCONF;
 	}
@@ -288,14 +314,27 @@ init_services()
 			if (buf[0] == '#')
 				continue;
 
+#ifdef STRICT
+#define checkbuf(b, cmd) (!strncasecmp(b, cmd, strlen(cmd)) && \
+			  (!b[strlen(cmd)] || isspace(b[strlen(cmd)])))
+#else
 #define checkbuf(b, cmd) (!strncasecmp(b, cmd, strlen(cmd)))
+#endif
 
+
+			for (cp=buf; cp && *cp && isspace(*cp); cp++);
+			if  (cp && *cp) {
+			strcpy(buf,cp);
 			if (checkbuf(buf, CMD_ORDER)) {
 				cp = strpbrk(buf, " \t");
-				if (!cp) 
+				if (!cp 
+#ifdef STRICT
+				    || !*(cp+1)
+#endif
+				    )
                                 {
                                     bad_config_format(CMD_ORDER);
-                                } else 
+                                } else
                                 {
                                     do {
 					while (*cp == ' ' || *cp == '\t')
@@ -311,63 +350,126 @@ init_services()
 					else
                                         {
                                             bad_config_format(CMD_ORDER);
+#if NLS
+			fprintf(stderr, catgets(_libc_cat, NetMiscSet,
+						NetMiscResolvInvalid,
+						"resolv+: \"%s\" is an invalid keyword\n"), cp);
+			fprintf(stderr, catgets(_libc_cat, NetMiscSet,
+						NetMiscResolvValid,
+						"resolv+: valid keywords are: %s, %s and %s\n"),
+#else
                                             fprintf(stderr, "resolv+: \"%s\" is an invalid keyword\n", cp);
                                             fprintf(stderr, "resolv+: valid keywords are: %s, %s and %s\n",
+#endif
                                                    ORD_BIND, ORD_HOSTS, ORD_NIS);
+
                                         }
-                                        
+
 					if (dp) cp = ++dp;
                                     } while (dp != NULL);
-                                    if (cc == 0) 
+                                    if (cc == 0)
                                     {
                                         bad_config_format(CMD_ORDER);
+#if NLS
+                                        fprintf(stderr,catgets(_libc_cat, NetMiscSet,
+							       NetMiscUnrecognized,
+							       "resolv+: search order not specified or unrecognized keyword, host resolution will fail.\n"));
+#else
                                         fprintf(stderr,"resolv+: search order not specified or unrecognized keyword, host resolution will fail.\n");
+#endif
                                     }
                                 }
-                                        
+
 			} else if (checkbuf(buf, CMD_HMA)) {
-                            if ( cp = strpbrk(buf, " \t") ) 
+#ifdef STRICT
+
+#if NLS
+#define check_legal(key,val) { \
+				for (cp=strpbrk(buf, " \t"); \
+				    cp && *cp && isspace(*cp); cp++); \
+				if (cp && *cp) {\
+				if (strlen(CMD_ON) == strlen(cp) && \
+				    checkbuf(cp,CMD_ON)) val = 1; \
+				else if (strlen(CMD_OFF) == strlen(cp) && \
+					 checkbuf(cp,CMD_OFF)) val = 0; \
+				else  { bad_config_format(key); \
+					fprintf(stderr, catgets(_libc_cat, NetMiscSet, \
+						NetMiscResolvInvalid, \
+						"resolv+: \"%s\" is an invalid keyword\n"), cp); }\
+				} else bad_config_format(key); }
+#else
+
+#define check_legal(key,val) { \
+				for (cp=strpbrk(buf, " \t"); \
+				    cp && *cp && isspace(*cp); cp++); \
+				if (cp && *cp) {\
+				if (strlen(CMD_ON) == strlen(cp) && \
+				    checkbuf(cp,CMD_ON)) val = 1; \
+				else if (strlen(CMD_OFF) == strlen(cp) && \
+					 checkbuf(cp,CMD_OFF)) val = 0; \
+				else  { bad_config_format(key); \
+                                     fprintf(stderr, "resolv+: \"%s\" is an invalid keyword\n", cp); } \
+				} else bad_config_format(key); }
+#endif /* NLS */
+			    check_legal(CMD_HMA,hosts_multiple_addrs);
+
+#else /* !STRICT */
+                            if ( cp = strpbrk(buf, " \t") )
                             {
 				while (*cp == ' ' || *cp == '\t') cp++;
 				if (checkbuf(cp, CMD_ON))
 					hosts_multiple_addrs = 1;
                             } else
                                 bad_config_format(CMD_HMA);
-                                
+#endif /* STRICT */
+
 			} else if (checkbuf(buf, CMD_SPOOF)) {
-                            if ( cp = strpbrk(buf, " \t") ) 
+#ifdef STRICT
+			    check_legal(CMD_SPOOF,spoof);
+#else
+                            if ( cp = strpbrk(buf, " \t") )
                             {
 				while (*cp == ' ' || *cp == '\t') cp++;
 				if (checkbuf(cp, CMD_ON))
 					spoof = 1;
                             } else
                                 bad_config_format(CMD_SPOOF);
+#endif
+
 
 			} else if (checkbuf(buf, CMD_SPOOFALERT)) {
-                            if ( cp = strpbrk(buf, " \t") ) 
+#ifdef STRICT
+			    check_legal(CMD_SPOOFALERT,spoofalert);
+#else
+                            if ( cp = strpbrk(buf, " \t") )
                             {
 				while (*cp == ' ' || *cp == '\t') cp++;
 				if (checkbuf(cp, CMD_ON))
 					spoofalert = 1;
                             } else
                                 bad_config_format(CMD_SPOOFALERT);
-                            
+#endif
+
 			} else if (checkbuf(buf, CMD_REORDER)) {
+#ifdef STRICT
+			    check_legal(CMD_REORDER,reorder);
+#else
 				if (cp = strpbrk(buf, " \t")) {
 				while (*cp == ' ' || *cp == '\t') cp++;
 				if (checkbuf(cp, CMD_ON))
 					reorder = 1;
 				} else
 					bad_config_format(CMD_REORDER);
+#endif
 
 			} else if (checkbuf(buf, CMD_TRIMDOMAIN)) {
 				if(numtrimdomains<MAXTRIMDOMAINS){
-                                    if ( cp = strpbrk(buf, " \t") ) 
+                                    if ( cp = strpbrk(buf, " \t") )
                                     {
 					while (*cp == ' ' || *cp == '\t') cp++;
-                                        if (cp) 
+                                        if (cp)
                                         {
-                                            (void) strcpy(tdp,cp);	
+                                            (void) strcpy(tdp,cp);
                                             trimdomain[numtrimdomains++]=tdp;
                                             tdp += strlen(cp)+1;
                                         } else
@@ -376,11 +478,23 @@ init_services()
                                         bad_config_format(CMD_TRIMDOMAIN);
 				}
 			}
+#ifdef STRICT
+				else {
+					if (cp = strpbrk(buf, " \t")) *cp='\0';
+#if NLS
+			fprintf(stderr, catgets(_libc_cat, NetMiscSet,
+						NetMiscResolvInvalid,
+						"resolv+: \"%s\" is an invalid keyword\n"), buf);
+#else
+					fprintf(stderr, "resolv+: \"%s\" is an invalid keyword\n", buf);
+#endif
+				}
+#endif
+			}
 		}
-		
 		service_order[cc] = SERVICE_NONE;
+		fclose(fd);
 	}
-	fclose(fd);
 	/* override service_order if environment variable */
 	if(NULL!=(cp=getenv(ENV_SERVORDER))){
 		cc=0;
@@ -453,7 +567,7 @@ init_services()
 			} while(cp=strtok(NULL, TOKEN_SEPARATORS));
 		}
 	}
-	
+
 	service_done = 1;
 }
 
@@ -644,7 +758,7 @@ gethostbyname(const char *name)
 #endif
 				return (&host);
 			}
-			if (!isdigit(*cp) && *cp != '.') 
+			if (!isdigit(*cp) && *cp != '.')
 				break;
 		}
 
@@ -665,7 +779,7 @@ gethostbyname(const char *name)
 			hp = getanswer(&buf, n, 0);
 			if (h_addr_ptrs[1] && reorder)
 				reorder_addrs(hp);
-			if (hp) 
+			if (hp)
 				return trim_domains(hp);
 			break;
 		case SERVICE_HOSTS:
@@ -700,13 +814,16 @@ gethostbyaddr(const char *addr, int len, int type)
 	register struct hostent *hp;
 	char qbuf[MAXDNAME];
 	extern struct hostent *_gethtbyaddr();
-	
+
 	if (type != AF_INET)
 		return ((struct hostent *) NULL);
 
 	if (!service_done)
 	  init_services();
 
+#if NLS
+	libc_nls_init();
+#endif
 	cc = 0;
 	while (service_order[cc] != SERVICE_NONE) {
 	        switch (service_order[cc]) {
@@ -729,35 +846,35 @@ gethostbyaddr(const char *addr, int len, int type)
 			if (hp) {
 				if(spoof){
 					/* Spoofing check code by
-					 * Caspar Dik <casper@fwi.uva.nl> 
+					 * Caspar Dik <casper@fwi.uva.nl>
 					 */
 					char nambuf[MAXDNAME+1];
 					int ntd, namelen = strlen(hp->h_name);
 					char **addrs;
-					
+
 					if (namelen >= MAXDNAME)
 						return (struct hostent *)NULL;
 					(void) strcpy(nambuf,hp->h_name);
 					nambuf[namelen] = '.';
 					nambuf[namelen+1] = '\0';
 
-					/* 
+					/*
 					 * turn off domain trimming,
-					 * call gethostbyname(), then turn	
+					 * call gethostbyname(), then turn
 					 * it back on, if applicable. This
 					 * prevents domain trimming from
 					 * making the name comparison fail.
 					 */
-					ntd=numtrimdomains; 
+					ntd=numtrimdomains;
 					numtrimdomains=0;
 					hp=gethostbyname(nambuf);
 					numtrimdomains=ntd;
 					nambuf[namelen] = 0;
 					/*
-					* the name must exist and the name 
-					* returned by gethostbyaddr must be 
-					* the canonical name and therefore 
-					* identical to the name returned by 
+					* the name must exist and the name
+					* returned by gethostbyaddr must be
+					* the canonical name and therefore
+					* identical to the name returned by
 					* gethostbyname()
 					*/
 					if (!hp || strcmp(nambuf, hp->h_name)){
@@ -768,7 +885,7 @@ gethostbyaddr(const char *addr, int len, int type)
 					* now check the addresses
 					*/
 #if defined(h_addr) || BSD >= 43
-					for (addrs = hp->h_addr_list; 
+					for (addrs = hp->h_addr_list;
 						*addrs; addrs++){
 						if (!bcmp(addrs[0], addr, len))
 							return trim_domains(hp);
@@ -783,13 +900,19 @@ gethostbyaddr(const char *addr, int len, int type)
 						openlog("resolv", LOG_PID,
 						    LOG_AUTH);
 						syslog(LOG_NOTICE,
+#if NLS
+						    catgets(_libc_cat, NetMiscSet,
+							    NetMiscPossibleSpoof,
+							    "gethostbyaddr: %s != %u.%u.%u.%u, possible spoof attempt"),
+#else
 						    "gethostbyaddr: %s != %u.%u.%u.%u, possible spoof attempt",
+#endif
 						    hp->h_name,
 						    ((unsigned)addr[0]&0xff),
 						    ((unsigned)addr[1]&0xff),
 						    ((unsigned)addr[2]&0xff),
 						    ((unsigned)addr[3]&0xff));
-					} 
+					}
 					return (struct hostent *)NULL;
 				}
 				hp->h_addrtype = type;
@@ -806,7 +929,7 @@ gethostbyaddr(const char *addr, int len, int type)
 			break;
 		case SERVICE_HOSTS:
 			hp = _gethtbyaddr(addr, len, type);
-			if (hp) 
+			if (hp)
 				return hp;
 			h_errno = HOST_NOT_FOUND;
 			break;
@@ -883,7 +1006,7 @@ again:
 	host.h_name = cp;
 	q = host.h_aliases = host_aliases;
 	cp = strpbrk(cp, " \t");
-	if (cp != NULL) 
+	if (cp != NULL)
 		*cp++ = '\0';
 	while (cp && *cp) {
 		if (*cp == ' ' || *cp == '\t') {
@@ -903,20 +1026,20 @@ again:
 /* if hosts_multiple_addrs set, then gethtbyname behaves as follows:
  *  - for hosts with multiple addresses, return all addresses, such that
  *  the first address is most likely to be one on the same net as the
- *  host we're running on, if one exists. 
+ *  host we're running on, if one exists.
  *  - like the dns version of gethostsbyname, the alias field is empty
  *  unless the name being looked up is an alias itself, at which point the
  *  alias field contains that name, and the name field contains the primary
  *  name of the host. Unlike dns, however, this behavior will still take place
- *  even if the alias applies only to one of the interfaces. 
- *  - determining a "local" address to put first is dependant on the netmask 
- *  being such that the least significant network bit is more significant 
- *  than any host bit. Only strange netmasks will violate this. 
+ *  even if the alias applies only to one of the interfaces.
+ *  - determining a "local" address to put first is dependant on the netmask
+ *  being such that the least significant network bit is more significant
+ *  than any host bit. Only strange netmasks will violate this.
  *  - we assume addresses fit into u_longs. That's quite internet specific.
- *  - if the host we're running on is not in the host file, the address 
+ *  - if the host we're running on is not in the host file, the address
  *  shuffling will not take place.
  *                     - John DiMarco <jdd@cdf.toronto.edu>
- */ 
+ */
 struct hostent *
 _gethtbyname(const char *name)
 {
@@ -934,7 +1057,7 @@ _gethtbyname(const char *name)
 	static struct hostent ht;
 	static char *aliases[MAXALIASES];
 	static char namebuf[MAXHOSTNAMELEN];
-	
+
 	hap = ht_addr_ptrs;
 	lhap = loc_addr_ptrs;
 	*hap = NULL;
@@ -952,11 +1075,11 @@ _gethtbyname(const char *name)
 
 	_sethtent(0);
 	while (p = _gethtent()) {
-		if (strcasecmp(p->h_name, name) == 0) 
+		if (strcasecmp(p->h_name, name) == 0)
 			found++;
-		else 
+		else
 			for (cp = p->h_aliases; *cp != 0; cp++)
-				if (strcasecmp(*cp, name) == 0){ 
+				if (strcasecmp(*cp, name) == 0){
 					found++;
 					aliases[0]=name;
 					(void) strcpy(namebuf, p->h_name);
@@ -1010,23 +1133,23 @@ _gethtbyname(const char *name)
 		return((struct hostent *)NULL);
 	}
 
-	ht.h_aliases = aliases; 
+	ht.h_aliases = aliases;
 	ht.h_name = namebuf;
 
-	/* shuffle addresses around to ensure one on same net as local host 
+	/* shuffle addresses around to ensure one on same net as local host
 	   is first, if exists */
 	{
 		/* "best" address is assumed to be the one with the greatest
 		   number of leftmost bits matching any of the addresses of
 		   the local host. This assumes a netmask in which all net
-		   bits precede host bits. Usually but not always a fair 
+		   bits precede host bits. Usually but not always a fair
 		   assumption. */
- 
+
 		/* portability alert: assumption: iaddr fits in u_long.
 		   This is really internet specific. */
 		int i,j, best=0;
 		u_long bestval = (u_long)~0;
-		
+
 		for(i=0;loc_addr_ptrs[i];i++){
 			for(j=0;ht_addr_ptrs[j];j++){
 				u_long t, l, h;
@@ -1034,7 +1157,7 @@ _gethtbyname(const char *name)
 				bcopy(loc_addr_ptrs[i], (char *)&t,
 					ht.h_length);
 				l=ntohl(t);
-				bcopy(ht_addr_ptrs[j], (char *)&t, 
+				bcopy(ht_addr_ptrs[j], (char *)&t,
 					ht.h_length);
 				t=l^h;
 
@@ -1053,7 +1176,7 @@ _gethtbyname(const char *name)
 			ht_addr_ptrs[best]=tmp;
 		}
 	}
-	
+
 	ht.h_addr_list = ht_addr_ptrs;
 	return (&ht);
 }
