@@ -23,6 +23,7 @@ static char sccsid[] = "@(#)ruserpass.c	5.5 (Berkeley) 6/27/88";
 #include <sys/stat.h>
 #include <utmp.h>
 #include <pwd.h>
+#include <netdb.h>
 
 static void renv(const char *host, char **aname, char **apass);
 static char * renvlook(const char *host);
@@ -35,11 +36,17 @@ static int token();
 void
 ruserpass(const char *host, char **aname, char **apass)
 {
-	static char namebuf [16];
+	static char namebuf [256];
+	struct hostent *hp;
+	char    name[256]; /* a temp name buffer to avoid overlayyed */
 
-	renv(host, aname, apass);
+	if (hp = gethostbyname (host))
+		strncpy(name,hp->h_name, sizeof(name) - 1);
+	else
+		strncpy(name,host, sizeof(name) - 1);
+	renv(name, aname, apass);
 	if (*aname == 0 || *apass == 0)
-		rnetrc(host, aname, apass);
+		rnetrc(name, aname, apass); /*rnetrc would call gethostbyname */
 	if (*aname == 0) {
 		*aname = getlogin();
 		if (*aname == NULL) {
@@ -147,6 +154,10 @@ static struct toktab {
 	0,		0
 };
 
+#if NLS
+#include "../nls/nl_types.h"
+#endif
+
 static void
 rnetrc(const char *host, char **aname, char **apass)
 {
@@ -154,7 +165,11 @@ rnetrc(const char *host, char **aname, char **apass)
 	int t;
 	struct stat stb;
 	extern int errno;
+	struct hostent *hp;
 
+#if NLS
+	libc_nls_init();
+#endif
 	hdir = getenv("HOME");
 	if (hdir == NULL)
 		hdir = ".";
@@ -173,8 +188,14 @@ next:
 		continue;
 
 	case MACHINE:
-		if (token() != ID || strcmp(host, tokval))
+		if (token() != ID)
 			continue;
+		if(hp = gethostbyname (tokval))
+			if (strcmp(host, hp->h_name))
+				continue;
+		else
+			if (strcmp(host, tokval))
+				continue;
 		while ((t = token()) && t != MACHINE) switch(t) {
 
 		case LOGIN:
@@ -190,8 +211,15 @@ next:
 		case PASSWD:
 			if (fstat(fileno(cfile), &stb) >= 0
 			    && (stb.st_mode & 077) != 0) {
+#if NLS
+	fprintf(stderr, "%s\n", catgets(_libc_cat, NetMiscSet,
+				NetMiscNetrcWrongPasswordMode,
+				"Error - .netrc file not correct mode.\n\
+				 Remove password or correct mode."));
+#else
 	fprintf(stderr, "Error - .netrc file not correct mode.\n");
 	fprintf(stderr, "Remove password or correct mode.\n");
+#endif
 				exit(1);
 			}
 			if (token() && *apass == 0) {
@@ -206,7 +234,15 @@ next:
 			(void) token();
 			break;
 		default:
+#if NLS
+	fprintf(stderr, "%s %s\n",
+		catgets(_libc_cat, NetMiscSet,
+			NetMiscUnknownNetrcOption,
+			"Unknown .netrc option"),
+		tokval);
+#else
 	fprintf(stderr, "Unknown .netrc option %s\n", tokval);
+#endif
 			break;
 		}
 		goto done;
