@@ -5,6 +5,7 @@
 #include <utmp.h>
 #include <unistd.h>
 #include <memory.h>
+#include <pwd.h>
 #include "pathnames.h"
 
 static char name[9];
@@ -30,7 +31,8 @@ char *getlogin(void)
 			if(!strncmp(ttyn + 5, ut.ut_line, sizeof(ut.ut_line))) {
 				strncpy(name, ut.ut_user, sizeof(ut.ut_user));
 				close(fd);
-				return name;
+				if (getpwnam(name))
+					return name;
 			}
 		}
 		close(fd);
